require 'pretty_proxy'
require 'equivalent-xml'
require 'rspec-html-matchers'
require 'zlib'
require 'uri'

shared_examples 'an reader method who encapsulate a mutable variable' do
  context 'when the return is changed' do
    it 'does not change the next return value' do
      instance = described_class.new(*new_args)
      first_return = instance.send reader_method_name
      if change_return.respond_to? :call
        change_return.call first_return
      else
        first_return.send change_return
      end
      second_return = instance.send reader_method_name

      expect(second_return).to_not eq first_return
    end
  end
end

describe PrettyProxy do

  def generate_html_for_test(base_url, hyperlinks)
    doc = <<-END
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
	<head>
    #{base_url ? "<base href='#{base_url}' />" : ''}
		<title>A title</title>
		<meta http-equiv="content-type" content="application/xhtml+xml; charset=UTF-8" />
	</head>
	<body>
    <a href="ARG_0" >a link </a>
		<p><a href="ARG_1" >other link</a></p>
		<div>
      <a href="ARG_2" >another link</a>
      <p><a href="ARG_3" >yet another link</a></p>
    </div>
		<div>
      <a href="ARG_4" >and yet another link</a>
      <p><a href="ARG_5" >almost there</a></p>
      <p><a href="ARG_6" >the last link</a></p>
    </div>
	</body>
</html>
    END
    doc.gsub!(/ARG_\d+/) { | match | hyperlinks[match[4..-1].to_i] }

    doc
  end

  let(:original_html_url) { 'http://site.net/p1' }
  let(:requested_to_proxy_url) do
    pp.proxify_hyperlink(original_html_url, 'http://site.net')
  end

  let (:link_examples) do
    [ 'http://site.net/p2/p2_2/', 'http://othersite.net', '../p3',
      '../p2/p2_2/', 'http://site.net/proxy/p1', '../proxy/p1', '/p1'
    ]
  end

  let(:html_base_href) { 'http://othersite.net/p1' }
  let(:original_html_with_base) do
    generate_html_for_test(html_base_href, link_examples)
  end
  let(:proxified_html_with_base) do
    # in this case only the first link is proxified because all others are
    # pointing for a site outside proxy control (or for the proxy itself)
    generate_html_for_test(html_base_href, [
      'http://site.net/proxy/p2/p2_2/', 'http://othersite.net',
      'http://othersite.net/p3', 'http://othersite.net/p2/p2_2/',
      'http://site.net/proxy/p1', 'http://othersite.net/proxy/p1',
      'http://othersite.net/p1'
    ])
  end

  let(:original_html) do
    generate_html_for_test(nil, link_examples)
  end
  let(:proxified_html) do
    generate_html_for_test(original_html_url, [
      'http://site.net/proxy/p2/p2_2/', 'http://othersite.net',
      'http://site.net/p3', 'http://site.net/proxy/p2/p2_2/',
      'http://site.net/proxy/p1', 'http://site.net/proxy/p1',
      'http://site.net/proxy/p1'
    ])
  end

  let (:correct_new_args_example) { ['/proxy/', 'http://myoriginalsite.com', '/content'] }
  let (:pp) { described_class.new(*correct_new_args_example) }

  describe '.new' do
    subject (:new) { described_class.method :new }

    [ {desc: 'accept original_paths as a String',
        args: ['/proxy/', 'http://myoriginalsite.com', '/content']},
      {desc: 'accept original_paths as an object who yelds strings with #each',
        args: ['/proxy/', 'http://myoriginalsite.com', ['/content', '/other_content']]},
      {desc: 'accept https in the original_domain',
        args: ['/proxy/', 'https://myoriginalsite.com', ['/content']]},
      {desc: 'accept port in the original_domain',
        args: ['/proxy/', 'https://myoriginalsite.com:8080', ['/content']]}
    ].each do | happy_case |
      it happy_case[:desc] do
        expect(new.call(*happy_case[:args])).to be_a_instance_of described_class
      end
    end

    let (:right_args) { correct_new_args_example }
    context "when proxy_path doesn't start with a '/'" do
      it { expect {new.call('proxy/', right_args[1], right_args[2])}.to raise_error(PrettyProxy::ConfigError) }
    end

    context "when proxy_path doesn't end with a '/'" do
      it { expect {new.call('/proxy', right_args[1], right_args[2])}.to raise_error(PrettyProxy::ConfigError) }
    end

    context 'when the original_domain is invalid' do
      it { expect {new.call(right_args[0], 'http://<invalid>.com', right_args[2])}.to raise_error(PrettyProxy::ConfigError)}
    end

    context 'when the original_domain has a query' do
      it { expect {new.call(right_args[0], 'http://myoriginalsite.com/?q=error', right_args[2])}.to raise_error(PrettyProxy::ConfigError)}
    end

    context 'when the original_domain has a fragment' do
      it { expect {new.call(right_args[0], 'http://myoriginalsite.com/#id', right_args[2])}.to raise_error(PrettyProxy::ConfigError)}
    end

    context "when the original_paths don't begin with a '/'" do
      it { expect {new.call(right_args[0], right_args[1], ['content'])}.to raise_error(PrettyProxy::ConfigError) }
    end
  end

  #NOTE: save ten lines of the not metaprogrammed way
  [:proxy_path, :original_domain, :original_paths].each do | reader_method |
    describe "##{reader_method.to_s}" do
      return_changers = { proxy_path: :chop!,
                          original_domain: ->(uri){ uri.host = 'otherdomain.com'},
                          original_paths: :clear }

      it_behaves_like 'an reader method who encapsulate a mutable variable' do
        let(:reader_method_name) { reader_method }
        let(:new_args) { ['/proxy/', 'http://myoriginalsite.com', '/content'] }
        let(:change_return) { return_changers[reader_method] }
      end
    end
  end
  
  # NOTE: excessive metaprogramming? only save 3~6 lines
  [ [ :proxy_path=, "when proxy_path doesn't end with a '/'", '/proxy'],
    [ :original_domain=, 'when the original_domain is invalid', 'http://<invalid>.com'],
    [ :original_paths=, "when the original_paths don't begin with a '/'", 'content']
  ].each do | error_case |
    writter, context_desc, invalid_input = *error_case
    describe "##{writter.to_s}" do
      context context_desc do
        it { expect {pp.send(writter, invalid_input)}.to raise_error(PrettyProxy::ConfigError) }
      end
    end
  end

  describe '#unproxify_url' do
    new_args = ['/proxys/sitez/', 'http://site.net', ['/p1', '/p2/p2_2/']]
    let (:pp) { described_class.new(*new_args) }

    context 'when the original_path has no trailing slash' do
      it 'allow no trailing slash in the url' do
        expect(pp.unproxify_url('http://myproxy.net/proxys/sitez/p1')).to eq 'http://site.net/p1'
      end
      it 'allow trailing slash in the url' do
        expect(pp.unproxify_url('http://myproxy.net/proxys/sitez/p1/')).to eq 'http://site.net/p1/'
      end
    end
    context 'when the original_path has a trailing slash' do
      it 'allow trailing slash in the url' do
        expect(pp.unproxify_url('http://myproxy.net/proxys/sitez/p2/p2_2/')).to eq 'http://site.net/p2/p2_2/'
      end
      it "don't allow no trailing slash" do
        expect { pp.unproxify_url('http://myproxy.net/proxys/sitez/p2/p2_2') }.to raise_error(PrettyProxy::ProxyError)
      end
    end
    it 'allow subdirectories inside that path' do
      expect(pp.unproxify_url('http://myproxy.net/proxys/sitez/p1/a/b/c/')).to eq 'http://site.net/p1/a/b/c/'
    end
    it 'preserve querys in the url' do
      expect(pp.unproxify_url('http://myproxy.net/proxys/sitez/p1/?q=error&l=pt')).to eq 'http://site.net/p1/?q=error&l=pt'
    end
    it 'preserve fragments in the url' do
      expect(pp.unproxify_url('http://myproxy.net/proxys/sitez/p1/#id')).to eq 'http://site.net/p1/#id'
    end
    it 'change the port to the original' do
      expect(URI.parse(pp.unproxify_url('http://myproxy.net:9292/proxys/sitez/p1/#id')).port).to eq 80
    end

    context 'when the url redirect to the own proxy' do
      let (:pp) { described_class.new('/', 'http://myoriginalsite.com/', '/content') }

      it { expect {pp.unproxify_url('http://myproxysite.com/proxy/proxy/')}.to raise_error(PrettyProxy::ProxyError) }
    end
    context "when the url don't begin with the proxy_path" do
      it { expect {pp.unproxify_url('http://myproxysite.com/no_proxy/content')}.to raise_error(PrettyProxy::ProxyError) }
    end
    context "when the proxy_path in the url isn't followed by a original_paths" do
      it { expect {pp.unproxify_url('http://myproxysite.com/proxy/other_content')}.to raise_error(PrettyProxy::ProxyError) }
    end
  end

  describe '#proxify_hyperlink' do
    let (:pp) { described_class.new('/proxy/', 'http://site.net', ['/p1', '/p2/p2_2/']) }

    it 'proxify urls that are inside the proxy control' do
      original_link = 'http://site.net/p2/p2_2/'
      url_proxy_page_with_the_link = 'http://theproxy.net/proxy/p1'
      proxified_link = 'http://theproxy.net/proxy/p2/p2_2/'
      expect(pp.proxify_hyperlink(original_link, url_proxy_page_with_the_link)).to eq proxified_link
    end
    it "don't proxify urls that are outside proxy control" do
      outside_site = 'http://othersite.net'
      proxy_url = 'http://theproxy.net/proxy/p1'
      expect(pp.proxify_hyperlink(outside_site, proxy_url)).to eq outside_site
    end

    context 'when the proxy itself is inside the proxy control' do
      let (:pp) { described_class.new('/proxy/', 'http://site.net', '/') }
      
      it "preserve urls to the proxy itself (don't double proxify)" do
        proxy_url = 'http://site.net/proxy/p1'
        another_proxy_url = 'http://site.net/proxy/p2/p2_2/'
        expect(pp.proxify_hyperlink(proxy_url, proxy_url)).to eq proxy_url
        expect(pp.proxify_hyperlink(proxy_url, another_proxy_url)).to eq proxy_url
      end
    end
  end

  describe '#proxify_html' do
    let (:pp) { described_class.new('/proxy/', 'http://site.net', ['/p1', '/p2/p2_2/']) }

    # valid mime_types are 'text/html' and 'application/xhtml+xml' (with any
    # others characters before or after)
    context 'when the mime_type is invalid' do
      it { expect {pp.proxify_html(original_html, 'http://site.net/proxy/p1', 'not a valid mime-type')}.to raise_error(described_class::ProxyError) }
    end

    context "when the content can't be parsed" do
      it 'return the original string' do
        # without the </b> to make this xml invalid the test fail
        # (the href is changed)
        page = '<a href="http://site.net/p2/p2_2/">test</a></b>'
        expect(pp.proxify_html(page, 'http://proxy.net/proxy/p1', 'application/xhtml+xml')).to equal(page)
      end
    end

    it 'apply #proxify_hyperlink in all anchors in the page' do
      # We aren't really testing with HTML, but with XHTML, what is a XML
      # This is because we dont have a matcher to test HTML equivalence, only XML equivalence
      # This test is not guaranteed to pass if the input is a HTML non-XHTML
      # The parse and unparse of the HTML can output a value who is not XML equivalent to the input
      # Maybe the way is use regex instead of Nokogiri to this work
      expect(pp.proxify_html(original_html, requested_to_proxy_url, 'application/xhtml+xml')).to be_equivalent_to(proxified_html)
    end

    context 'when the page has a base tag' do
      subject do
        pp.proxify_html(
          original_html_with_base,
          requested_to_proxy_url,
          'application/xhtml+xml'
        )
      end
      it 'do not alter the base tag' do
        should have_tag("base[href='#{html_base_href}']")
      end
      it 'use the base tag href as base url for relative links' do
        should be_equivalent_to(proxified_html_with_base)
      end
    end
  end

  describe '#rewrite_env' do
    # See http://rack.rubyforge.org/doc/SPEC.html for the rack env hash fields spec
    example_request = {'HTTP_HOST' => 'myproxysite.com',
                      'SCRIPT_NAME' => '',
                      'PATH_INFO' => '/proxy/content',
                      'QUERY_STRING' => '',
                      'SERVER_NAME' => 'myproxysite.com',
                      'SERVER_PORT' => '9292',
                      'rack.url_scheme' => 'http'}

    context "when the request is not prefixed by proxy_path" do
      let (:request_to_outside_content) { example_request.clone.update({'PATH_INFO' => '/no_proxy/content'}) }
      it { expect {pp.rewrite_env(request_to_outside_content)}.to raise_error(PrettyProxy::ProxyError) }
    end
    context "when the request don't point to a original_path" do
      let (:request_to_not_a_proxy) { example_request.clone.update({'PATH_INFO' => '/no_proxy/content'}) }
      it { expect {pp.rewrite_env(request_to_not_a_proxy)}.to raise_error(PrettyProxy::ProxyError) }
    end

    let (:by_proxy_request) { example_request.clone }
    let (:rewritten_env) { pp.rewrite_env by_proxy_request }

    context 'when the HTTP_HOST is not empty' do
      it 'change the HTTP_HOST and SERVER_NAME to the unproxyfied version' do
        expect(rewritten_env['HTTP_HOST']).to eq 'myoriginalsite.com'
        expect(rewritten_env['SERVER_NAME']).to eq 'myoriginalsite.com'
      end
    end
    context 'when the HTTP_HOST is empty' do
      let (:by_proxy_request) { t = example_request.clone; t.delete('HTTP_HOST'); t }
      it 'change the SERVER_NAME to the unproxyfied version' do
        expect(rewritten_env.has_key? 'HTTP_HOST').to be_false
        expect(rewritten_env['SERVER_NAME']).to eq 'myoriginalsite.com'
      end
    end
    context 'when the SCRIPT_NAME is not empty and the PATH_INFO is empty' do
      let (:by_proxy_request) { example_request.clone.update({'SCRIPT_NAME' => '/proxy/content',
                                                              'PATH_INFO' => ''}) }
      it 'changes only the SCRIPT_NAME' do
        expect(rewritten_env['SCRIPT_NAME']).to eq '/content'
        expect(rewritten_env['PATH_INFO']).to eq ''
      end
    end
    context 'when the PATH_INFO is not empty and the SCRIPT_NAME is empty' do
      it 'changes only the PATH_INFO' do
        expect(rewritten_env['PATH_INFO']).to eq '/content'
        expect(rewritten_env['SCRIPT_NAME']).to eq ''
      end
    end
    context 'when the SCRIPT_NAME and the PATH_INFO are not empty' do
      # NOTE: in a real request the SCRIPT_NAME have a trailing slash?
      #       even if the PATH_INFO start with a slash?
      let (:by_proxy_request) { example_request.update({'SCRIPT_NAME' => '/proxy',
                                                        'PATH_INFO' => '/content'}) }
      it 'change the SCRIPT_NAME to empty and the PATH_INFO has the full path' do
        expect(rewritten_env['PATH_INFO']).to eq '/content'
        expect(rewritten_env['SCRIPT_NAME']).to eq ''
      end
    end
  end

  describe '#rewrite_response' do
    let (:pp) { described_class.new('/proxy/', 'http://site.net', ['/p1', '/p2/p2_2/']) }
    # See http://rack.rubyforge.org/doc/SPEC.html for the rack env hash fields spec
    let (:original_env) do
      url = URI.parse(original_html_url)
      { 'HTTP_HOST' => url.host,
        'SCRIPT_NAME' => '',
        'PATH_INFO' => pp.proxy_path[0..-2] + url.path,
        'QUERY_STRING' => '',
        'SERVER_NAME' => url.host,
        'SERVER_PORT' => url.port,
        'rack.url_scheme' => url.scheme
      }
    end
    let (:rewritten_env) { pp.rewrite_env(original_env) }
    let (:response_example) do
      original_content = [200, {
        'content-type' => 'application/xhtml+xml',
        'content-encoding' => 'identity',
        'content-length' => original_html.bytesize.to_s
      }, [original_html]]
    end

    context 'when the content-type is html or xhtml' do
      let (:original_response) { response_example }
      subject { pp.rewrite_response(original_response, original_env, rewritten_env) }

      let (:rewritten_headers) { subject[1] }
      let (:rewritten_body) { subject[2].join }
      let (:requested_to_proxy_url) { Rack::Request.new(original_env).url }

      # NOTE: TESTING ONLY WITH XHTML, BY THE SAME MOTIVE EXPLAINED IN THE #proxify_html SPEC
      it 'apply #proxify_html to the body' do
        expect(rewritten_body).to be_equivalent_to pp.proxify_html(original_html, requested_to_proxy_url, 'application/xhtml+xml')
      end

      it 'change the content-length header to the new size of the body' do
        expect(rewritten_headers['content-length']).to eq rewritten_body.bytesize.to_s
      end

      context 'compressed with deflate' do
        it 'decompress, make the changes, and return it compressed again' do
          original_response[1].update({'content-encoding' => 'deflate'})
          deflate = Zlib::Deflate.method :deflate
          original_response[2] = [deflate.call(original_html)]
          inflate = Zlib::Inflate.method :inflate

          expect(inflate.call(rewritten_body)).to be_equivalent_to(proxified_html)
        end
      end

      context 'compressed with gzip' do
        it 'decompress, make the changes, and return it compressed again' do
          original_response[1].update({'content-encoding' => 'gzip'})
          gzip = ->(str) do
            return_str = ''
            gzip_stream = Zlib::GzipWriter.new(StringIO.new(return_str))
            gzip_stream.write str
            gzip_stream.close
            return_str
          end
          ungzip = ->(str) do
            Zlib::GzipReader.new(StringIO.new(str)).read
          end
          original_response[2] = [gzip.call(original_html)]

          expect(ungzip.call(rewritten_body)).to be_equivalent_to proxified_html
        end
      end
      
      context 'compressed with another method' do
        let (:original_response) { response_example[1].update({'content-encoding' => 'unknown-encoding'}); response_example }
        it { expect {subject}.to raise_error(PrettyProxy::ProxyError) }
      end
    end
  end
end

